// Experiments with tesselation and a geometry shader

extern texture2D Tex;
extern float Time;

sampler texsampler;

struct position_texture { float3 position: POSITION;  float2 texc: TEXTURECOORD; };
struct svposition_texture { float4 position: SV_POSITION;  float2 texc: TEXTURECOORD; };
struct tesselation_parameters { float edges[4]: SV_TessFactor;  float inside[2]: SV_InsideTessFactor; };

position_texture VERTEX_SHADER(position_texture input) { return input; }

tesselation_parameters HULL_SHADER_CONSTANT(InputPatch<position_texture, 4> inpatch, uint pid : SV_PrimitiveID)
{
	tesselation_parameters output;
	float subdivision = trunc(Time / 9) % 8 + 1;
	output.edges[0] = output.edges[1] = output.edges[2] = output.edges[3] = subdivision;
	output.inside[0] = output.inside[1] = subdivision;
	return output;
}

[domain("quad")][partitioning("integer")]  // integer, fractional_even, fractional_odd and pow2.
[outputtopology("triangle_cw")][outputcontrolpoints(4)][patchconstantfunc("HULL_SHADER_CONSTANT")]
position_texture HULL_SHADER(InputPatch<position_texture, 4> patch,
                             uint patch_index : SV_OutputControlPointID,
							 uint pid : SV_PrimitiveID)
{
	return patch[patch_index];
}

[domain("quad")]
position_texture DOMAIN_SHADER(tesselation_parameters input,
                                 float2 bary : SV_DomainLocation,
							     const OutputPatch<position_texture, 4> patch)
{
    position_texture Output;
    
    float3 topMidpoint = lerp(patch[0].position, patch[1].position, bary.x);
    float3 bottomMidpoint = lerp(patch[3].position, patch[2].position, bary.x);
    Output.position = lerp(topMidpoint, bottomMidpoint, bary.y);

    float2 topMidtexc = lerp(patch[0].texc, patch[1].texc, bary.x);
    float2 bottomMidtexc = lerp(patch[3].texc, patch[2].texc, bary.x);
    Output.texc = lerp(topMidtexc, bottomMidtexc, bary.y);

    return Output;    
}

[maxvertexcount(6)]
void GEOMETRY_SHADER(triangle position_texture input[3], inout TriangleStream<svposition_texture> TriStream )
{
    svposition_texture output;

	float2 offset = (input[0].position.xy + input[1].position.xy + input[2].position.xy) / 3;
	float phase = (trunc(Time) % 9 > 0 ? frac((Time - 1) / 9) * 9 / 8 * 6.28318 : 0);
	if (length(offset) > 0.25) phase = -phase;
	float2x2 rotation = float2x2(cos(phase * 2), -sin(phase * 2), sin(phase * 2), cos(phase * 2));

	for( int v=0; v<3; v++ )
	{
		float2 pos = mul(input[v].position.xy - offset, rotation) + offset * (1 - sin(phase));
		output.position = float4(pos, input[v].position.z, 1);
		output.texc = input[v].texc;
		TriStream.Append( output );
	}
	TriStream.RestartStrip();
}

float4 PIXEL_SHADER(svposition_texture input) : SV_Target
{
	return Tex.Sample(texsampler, input.texc);
}

technique11 Render
{
	pass P0
	{
		SetVertexShader( CompileShader( vs_5_0, VERTEX_SHADER() ) );
		SetHullShader( CompileShader( hs_5_0, HULL_SHADER() ) );
		SetDomainShader( CompileShader( ds_5_0, DOMAIN_SHADER() ) );
		SetGeometryShader( CompileShader( gs_5_0, GEOMETRY_SHADER() ) );
		SetPixelShader( CompileShader( ps_5_0, PIXEL_SHADER() ) );
	}
}
